"""nubison-model package."""

__version__ = "0.0.5"

from .Model import (
    ENV_VAR_MLFLOW_MODEL_URI,
    ENV_VAR_MLFLOW_TRACKING_URI,
    ModelContext,
    NubisonMLFlowModel,
    NubisonModel,
    register,
)
from .Service import build_inference_service, test_client

__all__ = [
    "ENV_VAR_MLFLOW_MODEL_URI",
    "ENV_VAR_MLFLOW_TRACKING_URI",
    "ModelContext",
    "NubisonModel",
    "NubisonMLFlowModel",
    "register",
    "build_inference_service",
    "test_client",
]
